/*
 * Rule
 * Copyright (C) 2011 by Apaq
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package dk.apaq.framework.criteria.jpa;

import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.Query;

import dk.apaq.framework.criteria.Criteria;
import dk.apaq.framework.criteria.Limit;
import dk.apaq.framework.criteria.Rule;
import dk.apaq.framework.criteria.Sorter;
import dk.apaq.framework.criteria.Sorter.SorterEntry;
import dk.apaq.framework.criteria.jpa.interpreters.CompareRuleInterpreter;
import dk.apaq.framework.criteria.jpa.interpreters.ContainsRuleInterpreter;
import dk.apaq.framework.criteria.jpa.interpreters.Interpreter;
import dk.apaq.framework.criteria.jpa.interpreters.IsNullRuleInterpreter;
import dk.apaq.framework.criteria.jpa.interpreters.LikeRuleInterpreter;
import dk.apaq.framework.criteria.rules.AndRule;
import dk.apaq.framework.criteria.rules.CompareRule;
import dk.apaq.framework.criteria.rules.ContainsRule;
import dk.apaq.framework.criteria.rules.IsNullRule;
import dk.apaq.framework.criteria.rules.LikeRule;
import dk.apaq.framework.criteria.rules.NotRule;
import dk.apaq.framework.criteria.rules.OrRule;

/**
 * Translates Rule, Sorters and Limits to JPA Query.
 */
public class CriteriaForJpa {

    private static final Map<Class<? extends Rule>, Interpreter<? extends Rule>> INTERPRETERMAP = 
        new HashMap<Class<? extends Rule>, Interpreter<? extends Rule>>();

    public static class WhereClause {

        private StringBuilder statement = new StringBuilder();
        private Map<String, Object> parameterMap = new HashMap<String, Object>();

        public void appendStatement(String jpql) {
            statement.append(jpql);
        }

        public void insertStatement(int index, String jpql) {
            statement.insert(index, jpql);
        }

        public void appendParameter(String parameter, Object value) {
            parameterMap.put(parameter, value);
        }

        public int getStatementLength() {
            return statement.length();
        }

        public String getStatement() {
            return statement.toString();
        }

        public Map<String, Object> getParameters() {
            return Collections.unmodifiableMap(parameterMap);
        }
    }

    static {
        INTERPRETERMAP.put(LikeRule.class, new LikeRuleInterpreter());
        INTERPRETERMAP.put(ContainsRule.class, new ContainsRuleInterpreter());
        INTERPRETERMAP.put(CompareRule.class, new CompareRuleInterpreter());
        INTERPRETERMAP.put(IsNullRule.class, new IsNullRuleInterpreter());
    }

    private CriteriaForJpa() { /* PRIVATE CONSTRUCTOR */ }
    
    public static void registerInterpreter(Class<? extends Rule> clazz, Interpreter<? extends Rule> interpreter) {
        INTERPRETERMAP.put(clazz, interpreter);
    }

    public static Query translate(EntityManager entityManager, Class entityClass, Criteria criteria) {
        return translate(entityManager, null, entityClass, criteria);
    }

    public static Query translate(EntityManager entityManager, String[] fields, Class entityClass, Criteria criteria) {
        Rule rule = criteria == null ? null : criteria.getRule();
        Limit limit = criteria == null ? null : criteria.getLimit();
        Sorter sorter = criteria == null ? null : criteria.getSorter();
        WhereClause whereClause = new WhereClause();
        boolean clauseAppended = buildWhereClause(whereClause, rule);

        StringBuilder jpqlBuilder = new StringBuilder();
        jpqlBuilder.append("select ");
        if (fields == null || fields.length == 0) {
            jpqlBuilder.append("e");
        } else {
            boolean firstField = true;
            for (String field : fields) {
                if (!firstField) {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append("e.");
                jpqlBuilder.append(field);
                firstField = false;
            }
        }

        jpqlBuilder.append(" from ");
        jpqlBuilder.append(entityClass.getSimpleName());
        jpqlBuilder.append(" e");

        if (clauseAppended) {
            jpqlBuilder.append(" WHERE ");
            jpqlBuilder.append(whereClause.getStatement());
        }

        if (sorter != null && !sorter.getSorterEntries().isEmpty()) {
            boolean firstOrderClause = true;
            for (SorterEntry entry : sorter.getSorterEntries()) {
                if (firstOrderClause) {
                    jpqlBuilder.append(" ORDER BY ");
                } else {
                    jpqlBuilder.append(", ");
                }
                jpqlBuilder.append(entry.getPropertyId());
                jpqlBuilder.append(entry.getDirection() == Sorter.Direction.Ascending ? " ASC" : " DESC");
                firstOrderClause = false;
            }
        }

        Query q = entityManager.createQuery(jpqlBuilder.toString());
        for (Map.Entry<String, Object> paramEntry : whereClause.getParameters().entrySet()) {
            q.setParameter(paramEntry.getKey(), paramEntry.getValue());
        }

        if (limit != null) {
            q.setFirstResult(limit.getOffset());
            q.setMaxResults(limit.getCount());
        }

        return q;

    }

    private static boolean buildWhereClause(WhereClause clause, Rule filter) {

        if (filter == null) {
            return false;
        }

        //First try an interpreter
        Interpreter interpreter = INTERPRETERMAP.get(filter.getClass());
        if (interpreter != null) {
            return interpreter.interpret(clause, filter);
        }


        //..else try our special handlers for filtercontainers.
        if (filter instanceof AndRule) {
            int currentRuleStartIndex = clause.getStatementLength();
            int count = 0;
            AndRule af = (AndRule) filter;

            boolean hasAddedRule = false;
            for (Rule currentRule : af.getRules()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentRule);

                if (statementAppended) {
                    if (hasAddedRule) {
                        clause.insertStatement(startIndex, " AND ");
                    }
                    hasAddedRule = true;
                    count++;
                }
            }

            if (hasAddedRule && count > 1) {
                clause.insertStatement(currentRuleStartIndex, "(");
                clause.appendStatement(")");
            }

            return hasAddedRule;
        }

        if (filter instanceof OrRule) {
            int currentRuleStartIndex = clause.getStatementLength();
            int count = 0;
            OrRule of = (OrRule) filter;
            boolean hasAddedRule = false;

            for (Rule currentRule : of.getRules()) {
                int startIndex = clause.getStatementLength();

                boolean statementAppended = buildWhereClause(clause, currentRule);

                if (statementAppended) {
                    if (hasAddedRule) {
                        clause.insertStatement(startIndex, " OR ");
                    }
                    hasAddedRule = true;
                    count++;
                }
            }

            if (hasAddedRule && count > 1) {
                clause.insertStatement(currentRuleStartIndex, "(");
                clause.appendStatement(")");
            }

            return hasAddedRule;
        }

        if (filter instanceof NotRule) {
            NotRule af = (NotRule) filter;

            //Special care of IsNull filter becauseof the special "IS [NOT] NULL" syntax
            if (af.getRuleElement() instanceof IsNullRule) {
                String propId = ((IsNullRule) af.getRuleElement()).getPropertyId();
                clause.appendStatement("e." + propId + " IS NOT NULL");
                return true;
            }

            int startIndex = clause.getStatementLength();
            boolean statementAppended = buildWhereClause(clause, af.getRuleElement());
            if (statementAppended) {
                clause.insertStatement(startIndex, "NOT (");
                clause.appendStatement(")");
                return true;
            } else {
                return false;
            }
        }

        return false;
    }
}
